import datetime
from urllib.parse import urlparse, urlunparse

ISO8601 = '%Y-%m-%dT%H:%M:%SZ'
COMPACT_ISO8601 = '%Y%m%dT%H%M%SZ'


def host_address(url):
    parsed = urlparse(url)
    hostname = f'{parsed.hostname}'
    if parsed.port:
        hostname = f'{parsed.hostname}:{parsed.port}'
    return (
        urlunparse((parsed.scheme, f'{hostname}', '', '', '', '')),
        parsed.username
    )


def iso_2_dt(s, fmt=ISO8601):
    dt = datetime.datetime.strptime(s, fmt)
    return dt.replace(tzinfo=datetime.timezone.utc)


def dt_2_iso(dt, fmt=ISO8601):
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=datetime.timezone.utc)
    dt = dt.astimezone(tz=datetime.timezone.utc)
    return dt.strftime(fmt)


def ciso_2_dt(s):
    return iso_2_dt(s, fmt=COMPACT_ISO8601)


def dt_2_ciso(dt):
    return dt_2_iso(dt, fmt=COMPACT_ISO8601)


def now():
    return datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)


def now_iso():
    return dt_2_iso(now())



class CCError(Exception):
    def __init__(self, message=None):
        msg = message or self.__class__.__name__
        super().__init__(msg)
        if isinstance(msg, (str, bytes)):
            self.messages = [msg]
        else:
            self.messages = msg


class InvalidWalletError(CCError):
    pass


class InvalidKeyError(InvalidWalletError):
    pass


class NoPrivateKeyError(InvalidWalletError):
    pass


class InvalidTransactionError(CCError):
    pass


class InvalidTransactionIdError(InvalidTransactionError):
    pass


class InvalidSignatureError(InvalidTransactionError):
    pass


class FutureTransactionError(InvalidTransactionError):
    pass


class ExpiredTransactionError(InvalidTransactionError):
    pass


class OutOfOrderTransactionError(InvalidTransactionError):
    pass


class UnsealedTransactionError(InvalidTransactionError):
    pass


class MissingWalletError(InvalidTransactionError):
    pass


class InsufficientFundsError(InvalidTransactionError):
    pass


class ImbalancedTransactionError(InvalidTransactionError):
    pass


class MissingInflowOutflowError(InvalidTransactionError):
    pass


class InvalidInflowOutflowError(InvalidTransactionError):
    pass


class InflowOutflowAddressMismatchError(InvalidTransactionError):
    pass


class SpentTransactionError(InvalidTransactionError):
    pass


class InvalidCoinbaseError(InvalidTransactionError):
    pass


class InvalidCoinbaseErrorRewardError(InvalidCoinbaseError):
    pass


class InvalidBlockError(CCError):
    pass


class InvalidBlockHashError(InvalidBlockError):
    pass


class InvalidPreviousHashError(InvalidBlockError):
    pass


class InvalidMerkleRootError(InvalidBlockError):
    pass


class MissingCoinbaseError(InvalidBlockError):
    pass


class SealedBlockError(InvalidBlockError):
    pass


class UnlinkedBlockError(InvalidBlockError):
    pass


class FutureBlockError(InvalidBlockError):
    pass


class InvalidProofError(InvalidBlockError):
    pass


class OutOfOrderBlockError(InvalidBlockError):
    pass


class InvalidBlockIndexError(InvalidBlockError):
    pass


class InvalidTargetError(InvalidBlockError):
    pass


class MissingBlockError(InvalidBlockError):
    pass


class InvalidChainError(CCError):
    pass


class EmptyChainError(InvalidChainError):
    pass


class MissingPreviousBlockError(InvalidChainError):
    pass



# Copyright 2023 Thomas Bohmbach, Jr.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import click
from flask import Flask
from flask.cli import FlaskGroup

__version__ = "1.2.0"


def create_app(app=None, register_browser=True, test_config=None):
    from .application import init_app
    from .cache import cache
    from .config import EnvAppSettings
    from .database import db
    from .tasks import init_tasks

    app = app or Flask(__name__)
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    if not test_config:
        app.config.from_object(EnvAppSettings.from_env())
        app.config.from_envvar('CANCELCHAIN_SETTINGS', silent=True)
    else:
        app.config.from_object(EnvAppSettings())
        app.config.from_mapping(test_config)

    init_app(app, register_browser=register_browser)
    try:
        db.init_app(app)
    except RuntimeError as e:
        app.logger.error(e)
    cache.init_app(app)
    init_tasks(app)

    @app.shell_context_processor
    def make_shell_context():
        return {'app': app, 'db': db}

    return app


@click.version_option(package_name='cancelchain')
@click.group(cls=FlaskGroup, create_app=create_app, add_version_option=False)
def cli():
    pass



import json
from dataclasses import dataclass, field
from functools import total_ordering

from cancelchain.block import Block
from cancelchain.exceptions import (
    EmptyChainError,
    FutureBlockError,
    ImbalancedTransactionError,
    InflowOutflowAddressMismatchError,
    InsufficientFundsError,
    InvalidBlockError,
    InvalidBlockIndexError,
    InvalidChainError,
    InvalidCoinbaseErrorRewardError,
    InvalidInflowOutflowError,
    InvalidPreviousHashError,
    InvalidTargetError,
    MissingInflowOutflowError,
    MissingPreviousBlockError,
    OutOfOrderBlockError,
    SpentTransactionError,
)
from cancelchain.milling import mill_hash_str
from cancelchain.models import BlockDAO, ChainDAO
from cancelchain.payload import Inflow, Outflow
from cancelchain.transaction import Transaction
from cancelchain.util import dt_2_iso, now

CURMUDGEON_PER_GRUMBLE = 100
GENESIS_HASH = mill_hash_str('GENESIS')
MAX_TARGET = '0' * 6 + 'F' * 58
REWARD = 100 * CURMUDGEON_PER_GRUMBLE
TARGET_GOAL_SECONDS = 600
TARGET_INTERVAL = 2016
TARGET_INTERVAL_SECONDS = TARGET_GOAL_SECONDS * TARGET_INTERVAL


def is_genesis_block(block):
    return block.prev_hash == GENESIS_HASH


@total_ordering
@dataclass()
class Chain:
    cid: int = field(default=None, compare=False)
    block_hash: str = field(default=None, compare=True)

    @property
    def blocks(self):
        return self.block_chain(block_hash=self.block_hash)

    @property
    def last_block(self):
        return next(self.blocks, None)

    @property
    def length(self):
        return self.last_block.idx + 1

    @property
    def target(self):
        return self.block_target()

    def block_chain(self, block=None, block_hash=None):
        if not block and block_hash:
            block = Block.from_db(block_hash)
        while block is not None:
            yield block
            prev_block = Block.from_db(block.prev_hash)
            if prev_block is None and not is_genesis_block(block):
                raise MissingPreviousBlockError()
            if is_genesis_block(block) and prev_block is not None:
                raise InvalidBlockError()
            block = prev_block

    def get_block_by_reverse_index(self, i=0):
        chain_dao = self.to_dao()
        if chain_dao is not None:
            index = self.last_block.idx - i
            block_dao = chain_dao.get_block(idx=index)
            return Block.from_dao(block_dao) if block_dao else None
        else:
            for index, block in enumerate(self.blocks):
                if index == i:
                    return block
            return None

    def block_target(self, block=None):
        if not self.last_block:
            return MAX_TARGET
        last_index = self.last_block.idx
        index = block.idx if block else last_index + 1
        if index == 0:
            return MAX_TARGET
        i = last_index - index
        prev_block = self.get_block_by_reverse_index(i + 1)
        prev_target = prev_block.target
        if index % TARGET_INTERVAL == 0:
            start_i = i + TARGET_INTERVAL
            start_block = self.get_block_by_reverse_index(start_i)
            interval_delta = prev_block.timestamp_dt - start_block.timestamp_dt
            factor = interval_delta.total_seconds() / TARGET_INTERVAL_SECONDS
            factor = min(max(factor, 0.25), 4.0)
            new_target = f"{int(int(prev_target, 16) * factor):064x}"
            if int(new_target, 16) > int(MAX_TARGET, 16):
                new_target = MAX_TARGET
            return new_target
        else:
            return prev_target

    def block_reward(self, block=None):
        return REWARD

    def link_block(self, block):
        last_block = self.last_block
        index = last_block.idx + 1 if last_block else 0
        prev_hash = last_block.block_hash if last_block else GENESIS_HASH
        target = self.target or MAX_TARGET
        block.link(index, prev_hash, target)

    def seal_block(self, block, wallet):
        block.seal(wallet, self.block_reward(block))

    def add_block(self, block):
        self.validate_block(block)
        block.to_db()
        self.block_hash = block.block_hash

    def validate(self, progress=None):
        _progress_next = progress.next if progress else lambda n=1: None
        if not self.last_block:
            raise EmptyChainError()
        for block in self.blocks:
            try:
                self.validate_block(block)
                _progress_next(n=1)
            except InvalidBlockError as e:
                raise InvalidChainError({f'Block #{block.idx}': e.messages})
        return True

    def validate_block(self, block):
        block.validate()
        if block.timestamp_dt > now():
            raise FutureBlockError()
        prev_block = Block.from_db(block.prev_hash)
        if prev_block is None and not is_genesis_block(block):
            raise InvalidPreviousHashError()
        if prev_block and block.timestamp_dt < prev_block.timestamp_dt:
            raise OutOfOrderBlockError()
        prev_hash = prev_block.block_hash if prev_block else None
        if block.prev_hash != prev_hash and not is_genesis_block(block):
            raise InvalidPreviousHashError()
        prev_index = prev_block.idx if prev_block else -1
        if block.idx != prev_index + 1:
            raise InvalidBlockIndexError()
        if block.target != self.block_target(block=block):
            raise InvalidTargetError()
        for txn in block.regular_txns:
            self.validate_block_txn(block, txn)
        self.validate_block_coinbase(block)

    def validate_block_txn(self, block, txn, txn_in_block=True):
        # add inflow amounts
        subject_amounts = {}
        other_amounts = 0
        for i in txn.inflows:
            amount, subject = self.validate_txn_inflow(
                block, txn, i, txn_in_block=txn_in_block
            )
            if subject:
                subject_amount = subject_amounts.get(subject, 0)
                subject_amounts[subject] = subject_amount + amount
            else:
                other_amounts += amount
        # subtract outflow amounts
        for o in txn.outflows:
            if o.forgive:
                forgive_amount = subject_amounts.get(o.forgive, 0)
                subject_amounts[o.forgive] = forgive_amount - o.amount
            elif o.subject:
                subject_amount = subject_amounts.get(o.subject)
                if subject_amount and subject_amount > 0:
                    if o.amount > subject_amount:
                        subject_amounts[o.subject] = 0
                        other_amounts -= (o.amount - subject_amount)
                    else:
                        subject_amounts[o.subject] = subject_amount - o.amount
                else:
                    other_amounts -= o.amount
            else:
                other_amounts -= o.amount
        if other_amounts != 0:
            raise ImbalancedTransactionError()
        for _, amount in subject_amounts.items():
            if amount != 0:
                raise ImbalancedTransactionError()

    def validate_txn_inflow(self, block, txn, i, txn_in_block=True):
        # txn inflow's outflow exists
        ioflow = None
        if ioflow_txn := self.get_transaction(
            i.outflow_txid, start_block=block
        ):
            ioflow = ioflow_txn.get_outflow(i.outflow_idx)
        if not ioflow:
            raise MissingInflowOutflowError()
        # inflow's outflow can't be for forgiveness or support
        if ioflow.forgive is not None or ioflow.support is not None:
            raise InvalidInflowOutflowError()
        # inflow's outflow address equals the txn address
        address = ioflow.address if ioflow.address else ioflow_txn.address
        if address != txn.address:
            raise InflowOutflowAddressMismatchError()
        # txn inflow's outflow not already used in other inflow
        num_inflows = self.get_inflows_count(
            block, i.outflow_txid, i.outflow_idx
        )
        if (num_inflows > 1 or (num_inflows > 0 and not txn_in_block)):
            raise SpentTransactionError()
        return ioflow.amount, ioflow.subject

    def validate_block_coinbase(self, block):
        block.validate_coinbase()
        reward = self.block_reward(block)
        if block.coinbase.outflows[0].amount != reward:
            raise InvalidCoinbaseErrorRewardError()

    def get_block(self, block_hash):
        block_dao = self.to_dao().get_block(block_hash)
        return Block.from_dao(block_dao) if block_dao else None

    def get_transaction(self, txid, start_block=None):
        block = start_block or self.last_block
        while (block_dao := BlockDAO.get(block.block_hash)) is None:
            for txn in block.txns:
                if txn.txid == txid:
                    return txn
            block = Block.from_db(block.prev_hash)
            if block is None:
                return None
        if block_dao is not None:
            txn_dao = block_dao.get_transaction_in_chain(txid)
            return Transaction.from_dao(txn_dao) if txn_dao else None
        return None

    def get_inflows_count(self, start_block, outflow_txid, outflow_idx):
        i = 0
        block = start_block
        while (block_dao := BlockDAO.get(block.block_hash)) is None:
            for txn in block.txns:
                for inflow in txn.inflows:
                    if (
                        inflow.outflow_txid == outflow_txid and
                        inflow.outflow_idx == outflow_idx
                    ):
                        i += 1
            block = Block.from_db(block.prev_hash)
            if block is None:
                break
        if block_dao is not None:
            i += block_dao.inflows_in_chain_count(outflow_txid, outflow_idx)
        return i

    def unspent_outflows(self, address, limit=None, filter_pending=False):
        amount = 0
        outflow_daos = self.to_dao().unspent_outflows(
            address, filter_pending=filter_pending
        )
        for outflow_dao in outflow_daos:
            txn = Transaction.from_dao(outflow_dao.transaction)
            index = outflow_dao.idx
            outflow = txn.get_outflow(index=index)
            amount += outflow.amount
            yield (outflow_dao.txid, outflow_dao.idx, outflow)
            if limit is not None and amount >= limit:
                break

    def unforgiven_outflows(self, subject, filter_pending=False):
        outflow_daos = self.to_dao().unforgiven_outflows(
            subject, filter_pending=filter_pending
        )
        for outflow_dao in outflow_daos:
            txn = Transaction.from_dao(outflow_dao.transaction)
            index = outflow_dao.idx
            outflow = txn.get_outflow(index=index)
            yield (outflow_dao.txid, outflow_dao.idx, outflow)

    def unforgiven_address_outflows(
        self, address, subject, limit=None, filter_pending=False
    ):
        amount = 0
        outflow_daos = self.to_dao().unforgiven_outflows(
            subject, address=address, filter_pending=filter_pending
        )
        for outflow_dao in outflow_daos:
            txn = Transaction.from_dao(outflow_dao.transaction)
            index = outflow_dao.idx
            outflow = txn.get_outflow(index=index)
            amount += outflow.amount
            yield (outflow_dao.txid, outflow_dao.idx, outflow)
            if limit is not None and amount >= limit:
                break

    def balance(self, address):
        return int(self.to_dao().wallet_balance(address))

    def subject_balance(self, subject):
        return int(self.to_dao().subject_balance(subject))

    def subject_support(self, subject):
        return int(self.to_dao().subject_support(subject))

    def create_transfer(self, wallet, amount, dest_address):
        address = wallet.address
        balance = 0
        t = Transaction()
        unspent = self.unspent_outflows(
            address, limit=amount, filter_pending=True
        )
        for txid, index, outflow in unspent:
            balance += outflow.amount
            t.add_inflow(Inflow(outflow_txid=txid, outflow_idx=index))
        if balance < amount:
            raise InsufficientFundsError()
        t.add_outflow(Outflow(amount=amount, address=dest_address))
        if balance - amount:
            t.add_outflow(Outflow(amount=balance-amount, address=address))
        t.set_wallet(wallet)
        t.seal()
        return t

    def create_subject(
        self, wallet, amount, subject, outflows=None, timestamp=None
    ):
        address = wallet.address
        balance = 0
        t = Transaction()
        if timestamp is not None:
            t.timestamp = dt_2_iso(timestamp)
        if outflows is not None:
            for outflow in outflows:
                balance += outflow[2]
                t.add_inflow(
                    Inflow(outflow_txid=outflow[0], outflow_idx=outflow[1])
                )
        if balance < amount:
            unspent = self.unspent_outflows(
                address, limit=amount-balance, filter_pending=True
            )
            for txid, index, outflow in unspent:
                balance += outflow.amount
                t.add_inflow(Inflow(outflow_txid=txid, outflow_idx=index))
        if balance < amount:
            raise InsufficientFundsError()
        t.add_outflow(Outflow(amount=amount, subject=subject))
        if balance - amount:
            t.add_outflow(Outflow(amount=balance-amount, address=address))
        t.set_wallet(wallet)
        t.seal()
        return t

    def create_forgive(self, wallet, amount, subject):
        address = wallet.address
        balance = 0
        t = Transaction()
        unforgiven = self.unforgiven_address_outflows(
            address, subject, limit=amount, filter_pending=True
        )
        for txid, index, outflow in unforgiven:
            balance += outflow.amount
            t.add_inflow(Inflow(outflow_txid=txid, outflow_idx=index))
        if balance < amount:
            raise InsufficientFundsError()
        t.add_outflow(Outflow(amount=amount, forgive=subject))
        if balance - amount:
            t.add_outflow(Outflow(amount=balance-amount, subject=subject))
        t.set_wallet(wallet)
        t.seal()
        return t

    def create_support(
        self, wallet, amount, subject, outflows=None, timestamp=None
    ):
        address = wallet.address
        balance = 0
        t = Transaction()
        if timestamp is not None:
            t.timestamp = dt_2_iso(timestamp)
        if outflows is not None:
            for outflow in outflows:
                balance += outflow[2]
                t.add_inflow(
                    Inflow(outflow_txid=outflow[0], outflow_idx=outflow[1])
                )
        if balance < amount:
            unspent = self.unspent_outflows(
                address, limit=amount-balance, filter_pending=True
            )
            for txid, index, outflow in unspent:
                balance += outflow.amount
                t.add_inflow(Inflow(outflow_txid=txid, outflow_idx=index))
        if balance < amount:
            raise InsufficientFundsError()
        t.add_outflow(Outflow(amount=amount, support=subject))
        if balance - amount:
            t.add_outflow(Outflow(amount=balance-amount, address=address))
        t.set_wallet(wallet)
        t.seal()
        return t

    def to_dict(self):
        return {
            'cid': self.cid,
            'block_hash': self.block_hash
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_dao(self, create=False):
        dao = None
        dao = ChainDAO.get(block_hash=self.block_hash)
        if dao is None and self.cid is not None:
            dao = ChainDAO.get(id=self.cid)
            try:
                dao.set_block_hash(self.block_hash)
            except Exception:
                dao = None
        if not dao:
            dao = ChainDAO.get(block_hash=self.block_hash)
        if not dao and create:
            dao = ChainDAO(self.block_hash)
        return dao

    def to_db(self):
        dao = self.to_dao(create=True)
        dao.commit()
        self.cid = dao.id

    def __lt__(self, other):
        return self.length < other.length

    @classmethod
    def from_dict(cls, d):
        cid = d.get('cid')
        block_hash = d.get('block_hash')
        chain = cls(cid=cid, block_hash=block_hash)
        return chain

    @classmethod
    def from_json(cls, j):
        return cls.from_dict(json.loads(j))

    @classmethod
    def from_dao(cls, dao):
        return cls(cid=dao.id, block_hash=dao.block_hash)

    @classmethod
    def from_db(cls, cid=None, block_hash=None):
        if cid:
            dao = ChainDAO.get(id=cid)
        else:
            dao = ChainDAO.get(block_hash=block_hash)
        return cls.from_dao(dao) if dao else None



from dataclasses import asdict

from marshmallow import Schema, fields, post_dump, validate

from cancelchain.util import iso_2_dt
from cancelchain.wallet import (
    ADDRESS_TAG,
    Wallet,
    b58decode,
    b64decode,
    b64encode,
)


def asdict_sans_none(dc):
    return asdict(
        dc, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
    )


def validate_address(public_key_b64, address):
    wallet = Wallet(b64ks=public_key_b64)
    return (wallet is not None) and address == wallet.address


def validate_address_format(address):
    try:
        if (
            address.startswith(ADDRESS_TAG) and
            address.endswith(ADDRESS_TAG) and
            len(b58decode(
                address.removeprefix(ADDRESS_TAG).removesuffix(ADDRESS_TAG)
            )) == 32
        ):
            return True
    except Exception:
        pass
    return False


def validate_base64(s):
    try:
        return b64encode(b64decode(s)) == s
    except Exception:
        pass
    return False


def validate_public_key(public_key_b64):
    wallet = Wallet(b64ks=public_key_b64)
    return wallet is not None and wallet.private_key is None


def validate_signature(public_key_b64, signing_data, signature):
    wallet = Wallet(b64ks=public_key_b64)
    if wallet is not None:
        return wallet.validate_signature(signing_data, signature)
    return False


def validate_timestamp(s):
    try:
        _ = iso_2_dt(s)
        return True
    except Exception:
        pass
    return False


class Address(fields.String):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, validate_address_format)


class Base64(fields.String):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, validate_base64)


class MillHash(Base64):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, validate.Length(equal=64))


class Timestamp(fields.String):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, validate_timestamp)


class PublicKey(Base64):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, validate_public_key)


class SansNoneSchema(Schema):
    @post_dump
    def remove_none_values(self, data, **kwargs):
        return {k: v for k, v in data.items() if v is not None}



from collections.abc import MutableSet
from dataclasses import dataclass, field
from json import JSONDecodeError

from marshmallow import (
    ValidationError,
    fields,
    post_load,
    validate,
    validates_schema,
)

from cancelchain.exceptions import (
    InvalidSignatureError,
    InvalidTransactionError,
    InvalidTransactionIdError,
    MissingWalletError,
    UnsealedTransactionError,
)
from cancelchain.milling import mill_hash_str
from cancelchain.models import (
    InflowDAO,
    OutflowDAO,
    PendingIOflowDAO,
    PendingTxnDAO,
    TransactionDAO,
)
from cancelchain.payload import Inflow, InflowSchema, Outflow, OutflowSchema
from cancelchain.schema import (
    Address,
    Base64,
    MillHash,
    PublicKey,
    SansNoneSchema,
    Timestamp,
    asdict_sans_none,
    validate_address,
    validate_signature,
)
from cancelchain.util import dt_2_iso, iso_2_dt, now_iso

VERSION_1 = '1'
MAX_FLOWS = 50
ADDRESS_MISSMATCH_MSG = 'Address/public key mismatch'


class TransactionSchema(SansNoneSchema):
    timestamp = Timestamp(required=True)
    txid = MillHash(required=True)
    address = Address(required=True)
    public_key = PublicKey(required=True)
    signature = Base64(required=False)
    inflows = fields.List(
        fields.Nested(InflowSchema),
        required=True,
        validate=validate.Length(min=0, max=MAX_FLOWS)
    )
    outflows = fields.List(
        fields.Nested(OutflowSchema),
        required=True,
        validate=validate.Length(min=1, max=MAX_FLOWS)
    )
    version = fields.String(required=True, validate=validate.Equal(VERSION_1))

    @validates_schema
    def validate_pk_address(self, data, **kwargs):
        if not validate_address(data.get('public_key'), data.get('address')):
            raise ValidationError(ADDRESS_MISSMATCH_MSG)

    @post_load
    def make_transaction(self, data, **kwargs):
        return Transaction(**data)


class RegularTransactionSchema(TransactionSchema):
    inflows = fields.List(
        fields.Nested(InflowSchema),
        required=True,
        validate=validate.Length(min=1, max=MAX_FLOWS)
    )


class CoinbaseTransactionSchema(TransactionSchema):
    inflows = fields.List(
        fields.Nested(InflowSchema),
        required=True,
        validate=validate.Length(equal=0)
    )
    outflows = fields.List(
        fields.Nested(OutflowSchema),
        required=True,
        validate=validate.Length(min=1, max=4)
    )


@dataclass(order=True)
class Transaction:
    timestamp: str = field(default_factory=now_iso)
    txid: str = field(default=None)
    address: str = field(default=None, compare=False)
    public_key: str = field(default=None, compare=False, repr=False)
    signature: str = field(default=None, compare=False, repr=False)
    inflows: list[Inflow] = field(default_factory=list, compare=False)
    outflows: list[Outflow] = field(default_factory=list, compare=False)
    version: str = field(default=VERSION_1, compare=False, repr=False)

    @property
    def timestamp_dt(self):
        return iso_2_dt(self.timestamp) if self.timestamp else None

    @property
    def data_csv(self):
        return ','.join([
            str(self.timestamp),
            str(self.address),
            str(self.public_key),
            ','.join(i.data_csv for i in self.inflows),
            ','.join(o.data_csv for o in self.outflows),
            str(self.version)
        ])

    @property
    def is_sealed(self):
        return self.txid is not None

    @property
    def signing_data(self):
        return ','.join([self.data_csv, self.txid]).encode()

    @property
    def schadenfreude(self):
        return sum([o.schadenfreude for o in self.outflows])

    @property
    def grace(self):
        return sum([o.grace for o in self.outflows])

    @property
    def mudita(self):
        return sum([o.mudita for o in self.outflows])

    def set_wallet(self, wallet):
        self.wallet = wallet
        self.address = self.wallet.address
        self.public_key = self.wallet.public_key_b64

    def add_inflow(self, i):
        self.inflows.append(i)

    def get_inflow(self, index=0):
        try:
            return self.inflows[index]
        except IndexError:
            return None

    def add_outflow(self, o):
        self.outflows.append(o)

    def get_outflow(self, index=0):
        try:
            return self.outflows[index]
        except IndexError:
            return None

    def calculate_txid(self):
        return mill_hash_str(self.data_csv)

    def seal(self):
        self.txid = self.calculate_txid()

    def sign(self):
        if not self.is_sealed:
            raise UnsealedTransactionError()
        if not self.wallet:
            raise MissingWalletError()
        self.signature = self.wallet.sign(self.signing_data)

    def validate_txid(self):
        if self.txid != self.calculate_txid():
            raise InvalidTransactionIdError()

    def validate_signature(self):
        if not validate_signature(
            self.public_key, self.signing_data, self.signature
        ):
            raise InvalidSignatureError()

    def validate(self, coinbase=False):
        if coinbase:
            errors = CoinbaseTransactionSchema().validate(self.to_dict())
        else:
            errors = RegularTransactionSchema().validate(self.to_dict())
        if errors:
            raise InvalidTransactionError(errors)
        self.validate_signature()
        self.validate_txid()

    def validate_coinbase(self):
        self.validate(coinbase=True)

    def to_dict(self):
        return asdict_sans_none(self)

    def to_json(self):
        return TransactionSchema().dumps(self.to_dict())

    def to_dao(self):
        return TransactionDAO.get(self.txid) or TransactionDAO(
            self.txid, self.version, self.timestamp_dt,
            address=self.address,
            public_key=self.public_key,
            signature=self.signature,
            inflow_daos=[
                InflowDAO(
                    self.txid, idx, inflow.outflow_txid, inflow.outflow_idx
                ) for idx, inflow in enumerate(self.inflows)],
            outflow_daos=[
                OutflowDAO(
                    self.txid, idx, outflow.amount,
                    address=outflow.address,
                    subject=outflow.subject,
                    forgive=outflow.forgive,
                    support=outflow.support
                ) for idx, outflow in enumerate(self.outflows)
            ]
        )

    def to_db(self):
        self.to_dao().commit()

    def __hash__(self):
        return int(self.txid, 16)

    @classmethod
    def from_dict(cls, d):
        try:
            return TransactionSchema().load(d)
        except ValidationError as e:
            raise InvalidTransactionError(e.messages)

    @classmethod
    def from_json(cls, j):
        try:
            return TransactionSchema().loads(j)
        except JSONDecodeError as je:
            raise InvalidTransactionError(je.msg)
        except ValidationError as ve:
            raise InvalidTransactionError(ve.messages)

    @classmethod
    def from_dao(cls, dao):
        return cls(
            timestamp=dt_2_iso(dao.timestamp),
            txid=dao.txid,
            address=dao.address,
            public_key=dao.public_key,
            signature=dao.signature,
            inflows=[
                Inflow(
                    outflow_txid=inflow_dao.outflow_txid,
                    outflow_idx=inflow_dao.outflow_idx
                ) for inflow_dao in dao.inflows
            ],
            outflows=[
                Outflow(
                    amount=outflow_dao.amount,
                    address=outflow_dao.address,
                    subject=outflow_dao.subject,
                    forgive=outflow_dao.forgive,
                    support=outflow_dao.support
                ) for outflow_dao in dao.outflows
            ],
            version=dao.version
        )

    @classmethod
    def from_db(cls, txid):
        dao = TransactionDAO.get(txid)
        return cls.from_dao(dao) if dao else None

    @classmethod
    def coinbase(cls, wallet, reward, schadenfreude, grace, mudita):
        outflows = []
        if reward:
            outflows.append(
                Outflow(amount=reward, address=wallet.address)
            )
        if schadenfreude:
            outflows.append(
                Outflow(amount=schadenfreude, address=wallet.address)
            )
        if grace:
            outflows.append(
                Outflow(amount=grace, address=wallet.address)
            )
        if mudita:
            outflows.append(
                Outflow(amount=mudita, address=wallet.address)
            )
        cb = cls(outflows=outflows)
        cb.set_wallet(wallet)
        cb.seal()
        cb.sign()
        return cb


class PendingTxnSet(MutableSet):
    def __contains__(self, txn):
        return PendingTxnDAO.get(txn.txid) is not None

    def __iter__(self):
        return (
            Transaction.from_json(json_data) for json_data in self.query_json()
        )

    def __len__(self):
        return PendingTxnDAO.count()

    def add(self, txn):
        dao = PendingTxnDAO(
            txid=txn.txid, timestamp=txn.timestamp_dt,
            json_data=txn.to_json()
        )
        dao.commit()
        for inflow in txn.inflows:
            ioflow_txn_dao = TransactionDAO.get(inflow.outflow_txid)
            if ioflow_txn_dao is not None:
                ioflow_dao = ioflow_txn_dao.outflows[inflow.outflow_idx]
                if ioflow_dao is not None:
                    PendingIOflowDAO(
                        txid=txn.txid,
                        outflow_txid=inflow.outflow_txid,
                        outflow_idx=inflow.outflow_idx,
                        pending_txn=dao,
                        outflow=ioflow_dao
                    ).commit()

    def discard(self, txn):
        PendingTxnDAO.get(txn.txid).delete()

    def query_json(self, earliest=None, expired=None):
        return PendingTxnDAO.json_datas(earliest=earliest, expired=expired)



import os
from datetime import timedelta
from http.client import responses

import click
import requests
from flask import current_app
from flask.cli import AppGroup, with_appcontext
from humanfriendly import format_timespan
from millify import millify
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.prompt import Confirm
from rich.rule import Rule
from rich.table import Table
from rich.text import Text

from cancelchain.api_client import ApiClient
from cancelchain.block import Block
from cancelchain.chain import CURMUDGEON_PER_GRUMBLE
from cancelchain.console import console
from cancelchain.database import db
from cancelchain.miller import Miller
from cancelchain.node import Node
from cancelchain.payload import encode_subject
from cancelchain.transaction import Transaction
from cancelchain.util import host_address, now_iso
from cancelchain.wallet import Wallet

REFRESH_PER_SECOND = 8
CHAIN_MISMATCH_MSG = 'Chain/file mismatch'

def grumble_to_curmudgeons(grumble):
    return int(CURMUDGEON_PER_GRUMBLE * float(grumble))


def human_curmudgeons(curmudgeons):
    balance = int(curmudgeons) / CURMUDGEON_PER_GRUMBLE
    return f'{balance:.2f}'.rstrip('0').rstrip('.')


def human_bignum(num):
    return millify(num, precision=2, drop_nulls=False)


def human_timespan(secs):
    return format_timespan(secs)


def http_error_message(e):
    try:
        msg = e.response.json().get('error')
        if msg:
            if isinstance(msg, dict):
                return ','.join([f"{k} => {v}" for k, v in msg.items()])
            elif isinstance(msg, list):
                return ','.join(msg)
            else:
                return msg
        else:
            return e.response.text
    except (AttributeError, requests.exceptions.JSONDecodeError):
        return responses.get(e.response.status_code)


def host_api_client(host=None, wallet_file=None):
    if not host:
        host = current_app.config.get('DEFAULT_COMMAND_HOST')
    if wallet_file:
        wallet = Wallet.from_file(wallet_file)
    else:
        host, address = host_address(host)
        wallet = current_app.wallets.get(address)
    return ApiClient(
        host, wallet, timeout=current_app.config.get('API_CLIENT_TIMEOUT')
    )


def address_wallet(address, wallet_file=None):
    if wallet_file:
        wallet = Wallet.from_file(wallet_file)
    else:
        wallet = current_app.wallets.get(address)
    if wallet is None or address != wallet.address:
        msg = f"No wallet for {address}"
        raise Exception(msg)
    return wallet


def read_last_line(file):
    with open(file, 'rb') as f:
        try:
            f.seek(-2, os.SEEK_END)
            while f.read(1) != b'\n':
                f.seek(-2, os.SEEK_CUR)
        except OSError:  # catch OSError in case of one line file
            f.seek(0)
        return f.readline().decode()


class ProgressBar():
    def __init__(self, title, console=None, total=None, completed=0):
        self.progress = Progress(
            # TextColumn(title),
            BarColumn(),
            TextColumn('{task.completed}/{task.total}'),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            TextColumn('['),
            TimeElapsedColumn(),
            TextColumn(']'),
            console=console
        )
        self.task_id = self.progress.add_task(
            title, total=total, completed=completed
        )
        self.panel = Panel.fit(
            self.progress,
            title=title
        )
        self.live = Live(
            self.panel,
            console=console,
            refresh_per_second=REFRESH_PER_SECOND
        )

    @property
    def console(self):
        return self.live.console

    def next(self, n=1):
        self.progress.advance(self.task_id, advance=n)

    def __enter__(self):
        self.live.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.live.__exit__(exc_type, exc_val, exc_tb)


class BlockSyncProgress():
    def __init__(self, peer=None, console=None):
        self.find_progress = Progress(
            SpinnerColumn(spinner_name='aesthetic', style='none'),
            TextColumn('{task.completed} Blocks'),
            TextColumn('['),
            TimeElapsedColumn(),
            TextColumn(']')
        )
        self.find_task_id = self.find_progress.add_task('Finding', total=None)
        self.load_title_text = TextColumn('Waiting...')
        self.load_count_text = TextColumn('')
        self.load_progress = Progress(
            self.load_title_text,
            BarColumn(),
            self.load_count_text,
            TaskProgressColumn(),
            TimeRemainingColumn(),
            TextColumn('['),
            TimeElapsedColumn(),
            TextColumn(']')
        )
        self.load_task_id = self.load_progress.add_task(
            'Loading', total=None, start=False
        )
        self.finding_panel = Panel.fit(
            self.find_progress,
            title='Finding Blocks'
        )
        self.loading_panel = Panel.fit(
            self.load_progress,
            title='Loading Blocks',
            border_style='dim'
        )
        progress_table = Table.grid()
        progress_table.add_row(self.finding_panel, self.loading_panel)
        self.live = Live(
            progress_table,
            console=console,
            refresh_per_second=REFRESH_PER_SECOND
        )
        self.progress = self.find_progress
        self.task_id = self.find_task_id
        self.console.print(
            Rule(title=f'Synchronizing with peer [bold]{peer}', align='left')
        )

    @property
    def console(self):
        return self.live.console

    def next(self, n=1):
        self.progress.advance(self.task_id, advance=n)

    def complete_find(self):
        block_count = self.find_progress.tasks[0].completed
        self.find_progress.update(
            self.find_task_id, total=block_count
        )
        self.load_title_text.text_format = ''
        return block_count

    def switch(self):
        block_count = self.complete_find()
        self.loading_panel.border_style = 'none'
        self.load_count_text.text_format = '{task.completed}/{task.total}'
        self.load_progress.update(self.load_task_id, total=block_count)
        self.load_progress.start_task(self.load_task_id)
        self.progress = self.load_progress
        self.task_id = self.load_task_id

    def finish(self):
        self.complete_find()

    def __enter__(self):
        self.live.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.live.__exit__(exc_type, exc_val, exc_tb)


class MillingProgress():
    def __init__(self, console=None):
        self.block = None
        self.chain = None
        self.progress = Progress(
            SpinnerColumn(spinner_name='aesthetic', style='milling'),
            TextColumn('{task.fields[hash_count]}h @'),
            TextColumn('{task.fields[hps]} hps'),
            TextColumn('['),
            TimeElapsedColumn(),
            TextColumn(']')
        )
        self.task_id = self.progress.add_task(
            'Milling', total=None, hash_count=0, hps=0
        )
        self.task = self.progress.tasks[0]
        self.panel = Panel.fit(
            self.progress,
            title="Milling",
            border_style='milling'
        )
        self.live = Live(
            self.panel,
            console=console,
            refresh_per_second=REFRESH_PER_SECOND
        )

    @property
    def hash_count(self):
        return str(human_bignum(self.task.completed))

    @property
    def hps(self):
        if self.task.elapsed:
            return human_bignum(self.task.completed / self.task.elapsed)
        else:
            return human_bignum(0)

    @property
    def console(self):
        return self.live.console

    @property
    def elapsed(self):
        if self.task.elapsed is None:
            return Text("-:--:--", style="progress.elapsed")
        delta = timedelta(seconds=int(self.task.elapsed))
        return Text(str(delta), style="progress.elapsed")

    def next(self, n=1):
        self.progress.update(
            self.task_id, advance=n, hash_count=self.hash_count, hps=self.hps
        )

    def next_block(self, block, chain):
        self.block = block
        self.chain = chain
        self.progress.reset(self.task_id)

    def __enter__(self):
        self.live.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.live.__exit__(exc_type, exc_val, exc_tb)

    def print_start(self):
        start_table = Table(show_header=False, border_style='milling')
        start_table.add_column('key', justify='right')
        start_table.add_column('value', justify='left')
        start_table.add_row('Block', str(self.block.idx))
        start_table.add_row(
            'Chain', self.chain.block_hash if self.chain else 'GENESIS'
        )
        start_table.add_row('Target', self.block.target)
        start_table.add_row('Started', now_iso())
        self.console.print(start_table)

    def print_stop(self, milled_block):
        stop_table = Table(show_header=False)
        stop_table.add_column('key', justify='right')
        stop_table.add_column('value', justify='left')
        stop_table.add_row('Stopped', now_iso())
        stop_table.add_row('Elapsed', self.elapsed)
        stop_table.add_row(
            'Hashes', f'{self.hash_count} @ {self.hps} hps'
        )
        label_text = Text('POW')
        style = 'milling.milled'
        if milled_block:
            pofw = milled_block.proof_of_work
            value_text = Text(f'{pofw} ({human_bignum(pofw)})', style=style)
            stop_table.add_row(label_text, value_text)
            stop_table.add_row('Block', f'{milled_block.block_hash}')
        elif self.block.proof_of_work is not None:
            style = 'milling.close'
            value_text = Text('SCOOPED (but so close)', style=style)
            stop_table.add_row(label_text, value_text)
        else:
            style = 'milling.scooped'
            value_text = Text('SCOOPED', style=style)
            stop_table.add_row(label_text, value_text)
        stop_table.border_style = style
        self.console.print(stop_table)
        self.console.print(Rule(style=style))


@click.command('init', help='Initialize the database.')
@with_appcontext
def init_db_command():
    try:
        db.create_all()
        console.print('Initialized the database.', style='success')
    except Exception as e:
        console.print(f'Initialization failed: {e}', style='error')


@click.command(
    'sync',
    help="Synchronize the node's block chain to that of its peers."
)
@with_appcontext
def sync_blocks_command():
    try:
        node = Node(
            host=current_app.config['NODE_HOST'],
            peers=current_app.config['PEERS'],
            clients=current_app.clients,
            logger=current_app.logger
        )
        for latest_block, peer in node.request_latest_blocks():
            try:
                progress_bar = BlockSyncProgress(peer=peer, console=console)
                with progress_bar as progress:
                    node.fill_chain(latest_block, progress=progress)
                    progress.finish()
            except requests.HTTPError as e:
                console.print(
                    f'Synchronization failed: {http_error_message(e)}',
                    style='error'
                )
            except Exception:
                console.print_exception()
    except Exception as e:
        console.print(f'Synchronization failed: {e}', style='error')


@click.command('validate', help="Validate the node's block chain.")
@with_appcontext
def validate_chain_command():
    try:
        node = Node(logger=current_app.logger)
        lc = node.longest_chain
        progress_bar = ProgressBar(
            'Validating Chain', console=console, total=lc.length
        )
        with progress_bar as progress:
            lc.validate(progress=progress)
    except Exception as e:
        console.print(f'The block chain is invalid: {e}', style='error')


@click.command('export')
@click.argument('file', type=click.Path())
@with_appcontext
def export_blocks_command(file):
    """Export the block chain to file.

    \b
    FILE is the file path to export the blocks to.
    If the file already exists, it will be appended to.
    """
    try:
        node = Node(logger=current_app.logger)
        lc = node.longest_chain
        lc_dao = lc.to_dao()
        progress = None
        last_block = None
        append_blocks = False
        if os.path.isfile(file) and (last_line := read_last_line(file)):
            last_block = Block.from_json(last_line)
            if lc_dao.get_block(last_block.block_hash) is None:
                raise Exception(CHAIN_MISMATCH_MSG)
            append_blocks = True
        last_idx = last_block.idx if last_block is not None else -1
        progress_bar = ProgressBar(
            "Exporting Blocks",
            console=console,
            total=lc_dao.block.idx+1,
            completed=last_block.idx+1 if last_block is not None else 0
        )
        with open(
            file, 'a' if append_blocks else 'w', encoding='utf-8'
        ) as f, progress_bar as progress:
            block_dao = lc_dao.get_block(idx=last_idx+1)
            while block_dao is not None:
                block = Block.from_dao(block_dao)
                f.write(block.to_json())
                f.write('\n')
                progress.next()
                block_dao = lc_dao.next_block(block_dao)
    except Exception:
        console.print_exception()
        console.print('Export failed', style='error')


@click.command('import')
@click.argument('file', type=click.Path(exists=True))
@with_appcontext
def import_blocks_command(file):
    """Import the block chain from file.

    \b
    FILE is the file path from which to import the blocks.
    """
    try:
        node = Node(logger=current_app.logger)
        with open(file, encoding='utf-8') as f:
            progress_bar = ProgressBar(
                "Importing Blocks",
                console=console,
                total=sum(1 for line in f)
            )
        with open(file, encoding='utf-8') as f, progress_bar as progress:
            for line in f:
                block = Block.from_json(line)
                if Block.from_db(block.block_hash) is None:
                    node.add_block(block)
                progress.next()
    except Exception:
        console.print_exception()
        console.print('Import failed', style='error')


@click.command('mill')
@click.argument('address')
@click.option(
    '-m', '--multi',
    is_flag=True,
    default=False,
    help='Use python multiprocessing when calculating hashes.'
)
@click.option(
    '-r', '--rounds',
    default=1,
    help='Number of rounds of milling between new block checks. (default 1)'
)
@click.option(
    '-s', '--size', 'worksize',
    default=100000,
    help=(
        'Number of hashes to calculate per round '
        '(per CPU if multiprocessing is enabled) '
        '(default 100000)'
    )
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for milling coinbase rewards.'
)
@click.option(
    '-p', '--peer',
    default=None,
    help=(
        "Peer node to poll before checking for new blocks and transactions."
    )
)
@click.option(
    '-b', '--blocks',
    default=0,
    help='Stop after this many blocks. (default 0 (run forever))'
)
@with_appcontext
def mill_command(address, multi, rounds, worksize, wallet, peer, blocks):
    """Start a milling process.

    \b
    ADDRESS is the address to use for milling coinbase rewards.
    """
    milling_wallet = address_wallet(address, wallet_file=wallet)
    if peer is not None and current_app.clients.get(peer) is None:
        msg = f"Peer {peer} client not configured."
        raise Exception(msg)
    miller = Miller(
        host=current_app.config['NODE_HOST'],
        peers=current_app.config['PEERS'],
        clients=current_app.clients,
        logger=current_app.logger,
        milling_wallet=milling_wallet,
        milling_peer=peer
    )
    if peer:
        try:
            progress_bar = BlockSyncProgress(peer=peer, console=console)
            with progress_bar as progress:
                miller.poll_latest_blocks(progress=progress)
                progress.finish()
        except Exception:
            console.print_exception()
            db.session.rollback()
    block_count = 0
    milling_progress = MillingProgress(console=console)
    with milling_progress as progress:
        progress.console.print()
        progress.console.print(
            Rule(
                title=f'Milling as address [bold]{milling_wallet.address}',
                align='left',
                style='milling'
            )
        )
        while (not blocks) or (block_count < blocks):
            try:
                chain = miller.longest_chain
                block = miller.create_block()
                progress.next_block(block, chain)
                progress.print_start()
                try:
                    milled_block = miller.mill_block(
                        block,
                        mp=multi,
                        rounds=rounds,
                        worksize=worksize,
                        progress=progress
                    )
                finally:
                    block_count += 1
                progress.print_stop(milled_block)
            except Exception:
                progress.console.print_exception()
                db.session.rollback()


txn_cli = AppGroup('txn', help='Command group to create transactions.')


@txn_cli.command('transfer')
@click.argument('from_address')
@click.argument('amount', type=click.FLOAT)
@click.argument('to_address')
@click.option(
    '-t', '--txn-wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for transaction source.'
)
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config).'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth.'
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Assume "yes" as answer to all prompts and run non-interactively.'
)
@with_appcontext
def create_transfer(
    from_address, amount, to_address, txn_wallet, host, wallet, yes
):
    """Create and post a transfer transaction.

    \b
    FROM_ADDRESS is the transaction source address.
    AMOUNT is the amount (as a float) of CCG to transfer.
    TO_ADDRESS is the transaction destination address.
    """
    try:
        txn_wallet = address_wallet(from_address, wallet_file=txn_wallet)
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_transfer_transaction(
            txn_wallet.public_key_b64,
            grumble_to_curmudgeons(amount),
            to_address
        )
        txn = Transaction.from_json(r.text)
        if not (confirm := yes):
            console.print(f'Transfer transaction created: {txn.txid}')
            confirm = Confirm.ask(
                'Do you want to sign and post the transaction?'
            )
        if confirm:
            txn.set_wallet(txn_wallet)
            txn.sign()
            client.post_transaction(txn)
            console.print('Transfer created.', style='success')
        else:
            console.print('Transfer aborted.', style='error')
    except requests.HTTPError as e:
        console.print(
            f'Transfer failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Transfer failed: {e}', style='error')


@txn_cli.command('subject')
@click.argument('address')
@click.argument('amount', type=click.FLOAT)
@click.argument('subject')
@click.option(
    '-t', '--txn-wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for transaction source.'
)
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config).'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth.'
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Assume "yes" as answer to all prompts and run non-interactively.'
)
@with_appcontext
def create_subject(address, amount, subject, txn_wallet, host, wallet, yes):
    """Create a subject ("cancel") transaction.

    \b
    ADDRESS is the transaction source address.
    AMOUNT is the amount (as a float) of CCG to apply.
    SUBJECT is the raw (unencoded) subject string.
    """
    try:
        txn_wallet = address_wallet(address, wallet_file=txn_wallet)
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_subject_transaction(
            txn_wallet.public_key_b64, grumble_to_curmudgeons(amount), subject
        )
        txn = Transaction.from_json(r.text)
        if not (confirm := yes):
            console.print(f'Subject transaction created: {txn.txid}')
            confirm = Confirm.ask(
                'Do you want to sign and post the transaction?'
            )
        if confirm:
            txn.set_wallet(txn_wallet)
            txn.sign()
            client.post_transaction(txn)
            console.print(f'Subject created: {txn.txid}', style='success')
        else:
            console.print('Subject aborted.', style='error')
    except requests.HTTPError as e:
        console.print(
            f'Subject failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Subject failed: {e}', style='error')


@txn_cli.command('forgive')
@click.argument('address')
@click.argument('amount', type=click.FLOAT)
@click.argument('subject')
@click.option(
    '-t', '--txn-wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for transaction source.'
)
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config).'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth.'
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Assume "yes" as answer to all prompts and run non-interactively.'
)
@with_appcontext
def create_forgive(address, amount, subject, txn_wallet, host, wallet, yes):
    """Create a forgive transaction.

    \b
    ADDRESS is the transaction source address.
    AMOUNT is the amount (as a float) of CCG to apply.
    SUBJECT is the raw (unencoded) subject string.
    """
    try:
        txn_wallet = address_wallet(address, wallet_file=txn_wallet)
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_forgive_transaction(
            txn_wallet.public_key_b64, grumble_to_curmudgeons(amount), subject
        )
        txn = Transaction.from_json(r.text)
        if not (confirm := yes):
            console.print(f'Subject transaction created: {txn.txid}')
            confirm = Confirm.ask(
                'Do you want to sign and post the transaction?'
            )
        if confirm:
            txn.set_wallet(txn_wallet)
            txn.sign()
            client.post_transaction(txn)
            console.print(f'Forgive created: {txn.txid}', style='success')
        else:
            console.print('Forgive aborted.', style='error')
    except requests.HTTPError as e:
        console.print(
            f'Forgive failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Forgive failed: {e} ', style='error')


@txn_cli.command('support')
@click.argument('address')
@click.argument('amount', type=click.FLOAT)
@click.argument('subject')
@click.option(
    '-t', '--txn-wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for transaction source.'
)
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config).'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth.'
)
@click.option(
    '-y', '--yes',
    is_flag=True,
    default=False,
    help='Assume "yes" as answer to all prompts and run non-interactively.'
)
@with_appcontext
def create_support(address, amount, subject, txn_wallet, host, wallet, yes):
    """Create a subject support transaction.

    \b
    ADDRESS is the transaction source address.
    AMOUNT is the amount (as a float) of CCG to apply.
    SUBJECT is the raw (unencoded) subject string.
    """
    try:
        txn_wallet = address_wallet(address, wallet_file=txn_wallet)
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_support_transaction(
            txn_wallet.public_key_b64, grumble_to_curmudgeons(amount), subject
        )
        txn = Transaction.from_json(r.text)
        if not (confirm := yes):
            console.print(f'Support transaction created: {txn.txid}')
            confirm = Confirm.ask(
                'Do you want to sign and post the transaction?'
            )
        if confirm:
            txn.set_wallet(txn_wallet)
            txn.sign()
            client.post_transaction(txn)
            console.print(f'Support created: {txn.txid}', style='success')
        else:
            console.print('Support aborted.', style='error')
    except requests.HTTPError as e:
        console.print(
            f'Support failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Support failed: {e}', style='error')


wallet_cli = AppGroup('wallet', help='Command group to work with wallets.')


@wallet_cli.command('create')
@click.option(
    '-d', '--walletdir',
    type=click.Path(exists=True),
    default=None,
    help="Parent directory for the wallet file (default from app config)."
)
@with_appcontext
def create_wallet(walletdir):
    """Create a new wallet file."""
    walletdir = walletdir or current_app.config.get('WALLET_DIR')
    w = Wallet()
    filename = w.to_file(walletdir=walletdir)
    console.print(f'Created {filename}', style='success')


@wallet_cli.command('balance')
@click.argument('address')
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config).'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth.'
)
@with_appcontext
def wallet_balance(address, host, wallet):
    """Get the wallet balance in CCG for an address.

    \b
    ADDRESS is the wallet address.
    """
    try:
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_wallet_balance(address)
        balance = r.json().get('balance')
        console.print(f'{human_curmudgeons(balance)} CCG', style='success')
    except requests.HTTPError as e:
        console.print(
            f'Balance failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Balance failed: {e}', style='error')


subject_cli = AppGroup('subject', help='Command group to work with subjects.')


@subject_cli.command('balance')
@click.argument('subject')
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config).'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth.'
)
@with_appcontext
def subject_balance(subject, host, wallet):
    """Get the balance (i.e. subject transactions minus forgiveness
       transactions) in CCG for a subject.

    \b
    SUBJECT is the raw (unencoded) subject string.
    """
    try:
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_subject_balance(encode_subject(subject))
        balance = r.json().get('balance')
        console.print(f'{human_curmudgeons(balance)} CCG', style='success')
    except requests.HTTPError as e:
        console.print(
            f'Subject balance failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Subject balance failed: {e}', style='error')


@subject_cli.command('support')
@click.argument('subject')
@click.option(
    '-h', '--host',
    default=None,
    help='The API host to use (default from app config)'
)
@click.option(
    '-w', '--wallet',
    type=click.Path(exists=True),
    default=None,
    help='Wallet file to use for API auth'
)
@with_appcontext
def support_balance(subject, host, wallet):
    """Get the support total in CCG for a subject.

    \b
    SUBJECT is the raw (unencoded) subject string.
    """
    try:
        client = host_api_client(host=host, wallet_file=wallet)
        r = client.get_subject_support(encode_subject(subject))
        support = r.json().get('support')
        console.print(f'{human_curmudgeons(support)} CCG', style='success')
    except requests.HTTPError as e:
        console.print(
            f'Support balance failed: {http_error_message(e)}', style='error'
        )
    except Exception as e:
        console.print(f'Support balance failed: {e}', style='error')



import json
import os
from base64 import standard_b64decode, standard_b64encode

import base58check
import Crypto.Random
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.Hash import SHA384
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5

from cancelchain.exceptions import InvalidKeyError, NoPrivateKeyError
from cancelchain.milling import mill_hash_bin

ADDRESS_TAG = 'CC'
KEY_SIZE = 2048


def b58decode(s):
    return base58check.b58decode(s.encode())


def b58encode(b):
    return base58check.b58encode(b).decode()


def b64decode(s):
    return standard_b64decode(s.encode())


def b64encode(b):
    return standard_b64encode(b).decode()


def export_binary_key(key, passphrase=None):
    if passphrase is None:
        return key.export_key(format='DER')
    else:
        return key.export_key(
            format='DER', pkcs=8, passphrase=passphrase
        )


def import_key(ks, passphrase=None):
    try:
        return RSA.import_key(ks, passphrase=passphrase)
    except Exception:
        return None


def import_b58_key(ks, passphrase=None):
    try:
        return import_key(b58decode(ks), passphrase=passphrase)
    except Exception:
        return None


def import_b64_key(ks, passphrase=None):
    try:
        return import_key(b64decode(ks), passphrase=passphrase)
    except Exception:
        return None


class Wallet:
    def __init__(self, b64ks=None, b58ks=None, ks=None, passphrase=None):
        if b64ks is not None:
            self.key = import_b64_key(b64ks, passphrase=passphrase)
        elif b58ks is not None:
            self.key = import_b58_key(b58ks, passphrase=passphrase)
        elif ks is not None:
            self.key = import_key(ks, passphrase=passphrase)
        else:
            self.key = RSA.generate(KEY_SIZE)
        if not (self.key and self.key.size_in_bits() == KEY_SIZE):
            raise InvalidKeyError()

    @property
    def private_key(self):
        return self.key if self.key.has_private() else None

    @property
    def public_key(self):
        return self.private_key.public_key() if self.private_key else self.key

    @property
    def private_key_b58(self):
        return self.export_private_key_b58()

    @property
    def public_key_b64(self):
        return b64encode(export_binary_key(self.public_key))

    @property
    def address(self):
        aks = b58encode(mill_hash_bin(export_binary_key(self.public_key)))
        return f'{ADDRESS_TAG}{aks}{ADDRESS_TAG}'

    def export_private_key_pem(self, passphrase=None):
        if self.private_key is None:
            raise NoPrivateKeyError()
        return self.private_key.export_key(
            pkcs=1 if passphrase is None else 8,
            passphrase=passphrase,
            protection="scryptAndAES128-CBC"
        )

    def export_private_key_b58(self, passphrase=None):
        if self.private_key is None:
            raise NoPrivateKeyError()
        return b58encode(
            export_binary_key(self.private_key, passphrase=passphrase)
        )

    def sign(self, data):
        if self.private_key is None:
            raise NoPrivateKeyError()
        signer = PKCS1_v1_5.new(self.private_key)
        hasher = SHA384.new(data=data)
        return b64encode(signer.sign(hasher))

    def validate_signature(self, data, signature):
        if not (data and signature):
            return False
        verifier = PKCS1_v1_5.new(self.public_key)
        hasher = SHA384.new(data=data)
        return verifier.verify(hasher, b64decode(signature))

    def encrypt(self, data):
        session_key = Crypto.Random.get_random_bytes(16)
        cipher_rsa = PKCS1_OAEP.new(self.public_key)
        enc_session_key = cipher_rsa.encrypt(session_key)
        cipher_aes = AES.new(session_key, AES.MODE_EAX)
        ciphertext, tag = cipher_aes.encrypt_and_digest(data)
        return b64encode(b''.join(
            x for x in (enc_session_key, cipher_aes.nonce, tag, ciphertext))
        )

    def decrypt(self, msg):
        def msg_parts(key_size, msg):
            for n in (key_size, 16, 16):
                yield msg[:n]
                msg = msg[n:]
            yield msg
        if self.private_key is None:
            raise NoPrivateKeyError()
        part = msg_parts(self.private_key.size_in_bytes(), b64decode(msg))
        cipher_rsa = PKCS1_OAEP.new(self.private_key)
        session_key = cipher_rsa.decrypt(next(part))
        cipher_aes = AES.new(session_key, AES.MODE_EAX, next(part))
        tag = next(part)
        data = cipher_aes.decrypt_and_verify(next(part), tag)
        return data

    def to_dict(self):
        return {'private_key': self.private_key_b58}

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_file(self, walletdir=None, passphrase=None):
        filename = f"{self.address}.pem"
        if walletdir:
            filename = os.path.join(walletdir, filename)
        with open(filename, 'wb') as f:
            f.write(self.export_private_key_pem(passphrase=passphrase))
        return filename

    def __repr__(self):
        return f"Wallet({self.address})"

    def __eq__(self, other):
        return self.key == other.key

    @classmethod
    def from_dict(cls, wallet_dict):
        return cls(b58ks=wallet_dict.get('private_key'))

    @classmethod
    def from_json(cls, wallet_json):
        return cls.from_dict(json.loads(wallet_json))

    @classmethod
    def from_file(cls, filename, passphrase=None):
        with open(filename, 'rb') as f:
            return cls(ks=f.read(), passphrase=passphrase)



from flask import Blueprint, abort, current_app, render_template
from werkzeug.exceptions import HTTPException

from cancelchain.block import Block
from cancelchain.models import BlockDAO, ChainDAO, TransactionDAO
from cancelchain.node import Node
from cancelchain.transaction import Transaction


def longest_chain():
    return Node(logger=current_app.logger).longest_chain


blueprint = Blueprint('browser', __name__)


@blueprint.route('/')
def index_view():
    try:
        lc = longest_chain()
    except HTTPException as e:
        return e
    except Exception as e:
        current_app.logger.exception(e)
        return e
    return render_template(
        'index.html',
        title="Home",
        lc=lc
    )


@blueprint.route('/chains')
def chains_view():
    try:
        chains_page = ChainDAO.chains().paginate()
    except HTTPException as e:
        return e
    except Exception as e:
        current_app.logger.exception(e)
        return e
    return render_template(
        'chains.html',
        title="Chains",
        chains_page=chains_page
    )


@blueprint.route('/block')
@blueprint.route('/block/<mill_hash:block_hash>')
def block_view(block_hash=None):
    try:
        if block_hash is None:
            lc = longest_chain()
            block_hash = lc.last_block.block_hash if lc is not None else None
        block_dao = BlockDAO.get(block_hash=block_hash)
        if block_dao is None:
            abort(404)
        block = Block.from_dao(block_dao)
    except HTTPException as e:
        return e
    except Exception as e:
        current_app.logger.exception(e)
        return e
    return render_template(
        'block.html',
        title=f"Block #{block.idx}: {block.block_hash}",
        block=block,
        block_dao=block_dao
    )


@blueprint.route('/transaction/<mill_hash:txid>')
def transaction_view(txid):
    try:
        inflows = []
        inflow_total = 0
        outflows = []
        outflow_total = 0
        transaction_dao = TransactionDAO.get(txid)
        if transaction_dao is None:
            abort(404)
        transaction = Transaction.from_dao(transaction_dao)
        for inflow in transaction.inflows:
            transaction_dao = TransactionDAO.get(inflow.outflow_txid)
            ioflow_txn = Transaction.from_dao(transaction_dao)
            ioflow = ioflow_txn.get_outflow(inflow.outflow_idx)
            inflows.append((inflow, ioflow_txn, ioflow))
            inflow_total += ioflow.amount
        for outflow in transaction.outflows:
            outflows.append(outflow)
            outflow_total += outflow.amount
    except HTTPException as e:
        return e
    except Exception as e:
        current_app.logger.exception(e)
        return e
    return render_template(
        'transaction.html',
        title=f"Transaction: {transaction.txid}",
        transaction=transaction,
        transaction_dao=transaction_dao,
        inflows=inflows,
        inflow_total=inflow_total,
        outflows=outflows,
        outflow_total=outflow_total
    )



import logging
from time import sleep

import requests
from sqlalchemy.exc import SQLAlchemyError

from cancelchain.block import TXN_TIMEOUT, Block
from cancelchain.chain import Chain, is_genesis_block
from cancelchain.exceptions import (
    InvalidBlockError,
    InvalidBlockHashError,
    InvalidTransactionIdError,
    MissingBlockError,
)
from cancelchain.models import (
    ChainDAO,
    ChainFill,
    ChainFillBlock,
    rollback_sesssion,
)
from cancelchain.signals import new_block as new_block_signal
from cancelchain.transaction import PendingTxnSet, Transaction
from cancelchain.util import host_address, now


class Node():
    def __init__(self, host=None, peers=None, clients=None, logger=None):
        self.host = host
        self.peers = peers or []
        self.clients = clients or {}
        self.logger = logger or logging.getLogger(__name__)
        self.pending_txns = PendingTxnSet()

    @property
    def longest_chain(self):
        longest = ChainDAO.longest()
        return Chain.from_dao(longest) if longest else None

    def send_transaction(self, txn, visited_hosts=None):
        visited_hosts = visited_hosts or []
        if self.host:
            host, _ = host_address(self.host)
            visited_hosts.append(host)
        for peer in self.peers:
            host, address = host_address(peer)
            if host not in visited_hosts:
                try:
                    self.clients.get(peer).post_transaction(
                        txn, visited_hosts=visited_hosts
                    )
                except requests.RequestException as re:
                    self.logger.warning(re)
                except Exception as e:
                    self.logger.exception(e)

    def receive_transaction(
        self, txid, txn_json, visited_hosts=None, process=True
    ):
        added = False
        txn = Transaction.from_json(txn_json)
        if txid != txn.txid:
            raise InvalidTransactionIdError()
        txn.validate()
        if txn not in self.pending_txns:
            try:
                self.pending_txns.add(txn)
            except SQLAlchemyError:
                rollback_sesssion()
                if txn.txid not in self.pending_txns:
                    raise
            added = True
        if process:
            self.send_transaction(txn, visited_hosts=visited_hosts)
        return txn if added else None

    def discard_expired_pending_txns(self):
        expired_dt = now() - TXN_TIMEOUT
        for txn in self.pending_txns:
            if txn.timestamp_dt <= expired_dt:
                self.pending_txns.discard(txn)

    def send_block(self, block, visited_hosts=None):
        visited_hosts = visited_hosts or []
        if self.host:
            host, _ = host_address(self.host)
            visited_hosts.append(host)
        for peer in self.peers:
            host, address = host_address(peer)
            if host not in visited_hosts:
                try:
                    r = self.clients.get(peer).post_block(
                        block,
                        visited_hosts=visited_hosts,
                        raise_for_status=False
                    )
                    if r.status_code == 404:
                        self.fill_peer(peer, block)
                except requests.RequestException as re:
                    self.logger.warning(re)
                except Exception as e:
                    self.logger.exception(e)

    def receive_block(
        self, block_json, block_hash=None, visited_hosts=None, process=True
    ):
        block = Block.from_json(block_json)
        if block is None:
            raise InvalidBlockError()
        if block_hash is not None and block_hash != block.block_hash:
            raise InvalidBlockHashError()
        if Block.from_db(block.block_hash):
            return None
        block.validate()
        prev_hash = block.prev_hash
        if Block.from_db(prev_hash) is None and not is_genesis_block(block):
            raise MissingBlockError()
        if process:
            block = self.process_block(block, visited_hosts=visited_hosts)
        return block

    def process_block(self, block, visited_hosts=None):
        if Block.from_db(block.block_hash):
            return None
        if block := self.add_block(block):
            new_block_signal.send(self, block=block)
            self.send_block(block, visited_hosts=visited_hosts)
        return block

    def add_block(self, block):
        try:
            chain = Chain.from_db(block_hash=block.prev_hash)
            if chain:
                chain.add_block(block)
            else:
                chain = self.create_chain(block=block)
            chain.to_db()
        except SQLAlchemyError:
            rollback_sesssion()
            if not Block.from_db(block.block_hash):
                raise
            block = None
        return block

    def create_chain(self, block=None):
        block_hash = block.prev_hash if block else None
        chain = Chain(block_hash=block_hash)
        if block:
            chain.add_block(block)
        return chain

    def request_block(self, block_hash):
        for peer in self.peers:
            try:
                r = self.clients.get(peer).get_block(
                    block_hash=block_hash, raise_for_status=False
                )
                if r.status_code == 200:
                    return Block.from_json(r.text)
            except requests.RequestException as re:
                self.logger.error(re)
            except Exception as e:
                self.logger.exception(e)
        return None

    def request_latest_blocks(self, peer=None):
        peers = [peer] if peer is not None else self.peers
        for peer in peers:
            try:
                r = self.clients.get(peer).get_block()
                yield Block.from_json(r.text), peer
            except requests.RequestException as re:
                self.logger.error(re)
            except Exception as e:
                self.logger.exception(e)

    def fill_peer(self, peer, last_block):
        blocks = []
        accepted = False
        block = last_block
        try:
            visited_hosts = []
            if self.host:
                host, _ = host_address(self.host)
                visited_hosts.append(host)
            client = self.clients.get(peer)
            while not accepted:
                blocks.insert(0, block)
                block = Block.from_db(block.prev_hash)
                r = client.post_block(
                    block,
                    visited_hosts=visited_hosts,
                    raise_for_status=False
                )
                if r.status_code in [200, 201, 202]:
                    accepted = True
                if r.status_code != 404:
                    r.raise_for_status()
            for block in blocks:
                accepted = False
                delay = 0
                while not accepted:
                    r = client.post_block(
                        block,
                        visited_hosts=visited_hosts,
                        raise_for_status=False
                    )
                    if r.status_code in [200, 201, 202]:
                        accepted = True
                    else:
                        if r.status_code != 404:
                            r.raise_for_status()
                        sleep(delay)
                        delay += 1
                        if delay > 10:
                            r.raise_for_status()
        except Exception as e:
            self.logger.exception(e)

    def fill_chain(self, last_block, progress=None):
        progress_next = progress.next if progress else lambda n=1: None
        progress_switch = progress.switch if progress else lambda: None
        chain_fill = None
        try:
            if Block.from_db(last_block.block_hash):
                return True
            chain_fill = ChainFill()
            chain_fill.commit()
            ChainFillBlock(
                block_hash=last_block.block_hash,
                idx=last_block.idx,
                block_json=last_block.to_json(),
                chain_fill=chain_fill
            ).commit()
            progress_next()
            block = last_block
            while True:
                is_genesis = is_genesis_block(block)
                prev_hash = block.prev_hash
                if Block.from_db(prev_hash) or is_genesis:
                    break
                block = self.request_block(prev_hash)
                if not block:
                    self.logger.error(f'Block request failed: {prev_hash}')
                    return False
                progress_next()
                ChainFillBlock(
                    block_hash=block.block_hash,
                    idx=block.idx,
                    block_json=block.to_json(),
                    chain_fill=chain_fill
                ).commit()
            progress_switch()
            for chain_fill_block in chain_fill.blocks:
                block = Block.from_json(chain_fill_block.block_json)
                self.add_block(block)
                new_block_signal.send(self, block=block)
                progress_next()
            return True
        except Exception as e:
            self.logger.exception(e)
        finally:
            if chain_fill is not None:
                chain_fill.delete()
        return False



import requests
from celery import Celery

celery = Celery(__name__)


def init_tasks(app):
    celery.conf.update(app.config)

    class ContextTask(celery.Task):
        def __call__(self, *args, **kwargs):
            with app.app_context():
                return self.run(*args, **kwargs)

    celery.Task = ContextTask
    return celery


@celery.task()
def post_process(url, data, headers=None):
    r = requests.post(
        url,
        headers=headers,
        data=data,
        timeout=360
    )
    r.raise_for_status()



import datetime
import os

from werkzeug.routing import BaseConverter, ValidationError

from cancelchain import __version__, api, browser, command
from cancelchain.api_client import ApiClient
from cancelchain.payload import decode_subject, validate_subject
from cancelchain.schema import validate_address_format, validate_base64
from cancelchain.util import host_address
from cancelchain.wallet import Wallet


def init_app(app, register_browser=True):
    app.wallets = read_wallets(app)
    app.clients = create_clients(app)

    app.url_map.converters['address'] = AddressConverter
    app.url_map.converters['mill_hash'] = MillHashConverter
    app.url_map.converters['subject'] = SubjectConverter

    app.register_blueprint(api.blueprint, url_prefix='/api')
    if register_browser:
        app.register_blueprint(browser.blueprint, url_prefix='/')
    app.cli.add_command(command.init_db_command)
    app.cli.add_command(command.sync_blocks_command)
    app.cli.add_command(command.validate_chain_command)
    app.cli.add_command(command.export_blocks_command)
    app.cli.add_command(command.import_blocks_command)
    app.cli.add_command(command.mill_command)
    app.cli.add_command(command.txn_cli)
    app.cli.add_command(command.wallet_cli)
    app.cli.add_command(command.subject_cli)

    @app.context_processor
    def inject_cc_version():
        return {'cc_version': __version__}

    @app.template_filter('utc_datetime')
    def utc_datetime(value, fmt="%a %b %d %H:%M:%S %Z"):
        if value.tzinfo is None or value.tzinfo.utcoffset(value) is None:
            value = value.replace(tzinfo=datetime.timezone.utc)
        value = value.astimezone(datetime.timezone.utc)
        return value.strftime(fmt) if value is not None else None

    @app.template_filter('human_subject')
    def human_subject(value):
        return decode_subject(value) if value is not None else None


def read_wallets(app):
    walletdir = app.config.get('WALLET_DIR')
    wallets = {}
    if walletdir and os.path.isdir(walletdir):
        for dirpath, _, filenames in os.walk(walletdir):
            for filename in filenames:
                if filename.endswith('.pem'):
                    try:
                        w = Wallet.from_file(os.path.join(dirpath, filename))
                        wallets[w.address] = w
                    except Exception as e:
                        app.logger.error(
                            f'Error reading {os.path.join(dirpath, filename)}'
                        )
                        app.logger.exception(e)
    return wallets


def create_clients(app):
    clients = {}
    timeout = app.config.get('API_CLIENT_TIMEOUT')
    for peer in app.config.get('PEERS'):
        host, address = host_address(peer)
        if wallet := app.wallets.get(address):
            clients[peer] = ApiClient(peer, wallet, timeout=timeout)
        else:
            app.logger.warning(
                f'Peer client wallet {address} for {host} not found'
            )
    return clients


class AddressConverter(BaseConverter):
    def to_python(self, value):
        if not validate_address_format(value):
            raise ValidationError
        return value


class MillHashConverter(BaseConverter):
    def to_python(self, value):
        if len(value) != 64 or not validate_base64(value):
            raise ValidationError
        return value


class SubjectConverter(BaseConverter):
    def to_python(self, value):
        if not validate_subject(value):
            raise ValidationError
        return value



import json

import requests

from cancelchain.block import MAX_TRANSACTIONS, TXN_TIMEOUT, Block
from cancelchain.chain import Chain
from cancelchain.milling import milling_generator
from cancelchain.node import Node
from cancelchain.signals import txn_failed as txn_failed_signal
from cancelchain.util import host_address, now


class Miller(Node):
    def __init__(
        self, host=None, peers=None, clients=None, logger=None,
        milling_wallet=None, milling_peer=None
    ):
        super().__init__(
            host=host, peers=peers, clients=clients, logger=logger
        )
        self.milling_client = None
        self.milling_peer = milling_peer
        if self.milling_peer is not None:
            self.milling_client = self.clients.get(self.milling_peer)
        self.milling_wallet = milling_wallet
        self.pending_txns_generator = None

    def pending_txns_gen(self):
        last_call = None
        while True:
            call_dt = now()
            if self.milling_client is not None:
                visited_hosts = [self.milling_client.host]
                try:
                    r = self.milling_client.get_pending_transactions(
                        earliest=last_call
                    )
                    for txn_json in r.json():
                        if txid := txn_json.get('txid'):
                            self.receive_transaction(
                                txid, json.dumps(txn_json),
                                visited_hosts=visited_hosts
                            )
                except requests.RequestException as re:
                    self.logger.error(re)
                except Exception as e:
                    self.logger.exception(e)
            last_call = call_dt
            yield last_call

    def update_pending_txns(self):
        if self.pending_txns_generator is None:
            self.pending_txns_generator = self.pending_txns_gen()
        _ = next(self.pending_txns_generator)
        self.discard_expired_pending_txns()

    def pending_chain_txns(self, chain):
        expired_dt = now() - TXN_TIMEOUT
        for txn in self.pending_txns:
            if (
                txn.timestamp_dt > expired_dt and
                not chain.get_transaction(txn.txid)
            ):
                yield txn

    def create_block(self):
        chain = self.longest_chain or self.create_chain()
        block = Block()
        chain.link_block(block)
        i = 0
        discard_txns = []
        self.update_pending_txns()
        for txn in self.pending_chain_txns(chain):
            try:
                chain.validate_block_txn(block, txn, txn_in_block=False)
                block.add_txn(txn)
                i += 1
                if i >= MAX_TRANSACTIONS - 1:
                    break
            except Exception as e:
                discard_txns.append(txn)
                txn_failed_signal.send(self, txn=txn, e=e)
        for txn in discard_txns:
            self.pending_txns.discard(txn)
        chain.seal_block(block, self.milling_wallet)
        return block

    def poll_latest_blocks(self, progress=None):
        latest_blocks = self.request_latest_blocks(peer=self.milling_peer)
        for latest_block, peer in latest_blocks:
            if Block.from_db(latest_block.block_hash) is None:
                self.fill_chain(latest_block, progress=progress)
                host, _ = host_address(peer)
                self.send_block(latest_block, visited_hosts=[host])

    def mill_block(
        self, block, mp=False, rounds=None, worksize=None, progress=None
    ):
        solved_block = None
        chain = Chain.from_db(block_hash=block.prev_hash)
        for proof_of_work in milling_generator(
            block, mp=mp, rounds=rounds, worksize=worksize, progress=progress
        ):
            if self.milling_peer is not None:
                self.poll_latest_blocks()
            longest_chain = self.longest_chain
            if longest_chain is not None and (
                chain is None or chain < longest_chain
            ):
                break
            if proof_of_work is not None:
                solved_block = self.receive_block(block.to_json())
        return solved_block



from rich.console import Console
from rich.theme import Theme

theme = Theme({
    'error': 'bold red',
    'milling': 'cyan',
    'milling.close': 'yellow',
    'milling.milled': 'bold green',
    'milling.scooped': 'dim',
    'success': 'bold green',
    # Overridden defaults
    'progress.elapsed': 'none',
    'progress.percentage': 'none',
    'progress.remaining': 'none',
    'rule.line': 'none'
})

console = Console(theme=theme, highlight=False)
